#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "vbe.h"
#include "vesamode.h"
#include "lrmi.h"

#include "list.h"
#include "module.h"

char *
snip (char *string)
{
  int i;

  /* This is always a 13 character buffer */
  /* and it's not always terminated. */
  string[12] = '\0';

  while (((i = strlen (string)) > 0) &&
	 (isspace (string[i - 1]) ||
	  (string[i - 1] == '\n') || (string[i - 1] == '\r')))
  {
    string[i - 1] = '\0';
  }
  return string;
}

struct node *
probe (struct node *devices, int params)
{
  struct vbe_info *vbe_info = NULL;
  struct vbe_edid1_info *edid_info = NULL;
  u_int16_t *mode_list = NULL;
  char manufacturer[4];
  int i;
  struct node *curdev;
  char buf[0x100];

  assert (sizeof (struct vbe_info) == 512);
  assert (sizeof (struct vbe_edid1_info) == 256);
  assert (sizeof (struct vbe_edid_detailed_timing) == 18);
  assert (sizeof (struct vbe_edid_monitor_descriptor) == 18);

  for (curdev=devices;curdev!=NULL;curdev=curdev->next)
  {
    if (strcmp(nodegetprop(curdev,"CLASS"),"VIDEO")==0)
    {


  vbe_info = vbe_get_vbe_info ();
  if (vbe_info == NULL)
  {
    return devices;
  }

  /* Signature. */
  sprintf (buf,"%c%c%c%c %d.%d",
	  vbe_info->signature[0], vbe_info->signature[1],
	  vbe_info->signature[2], vbe_info->signature[3],
	  vbe_info->version[1], vbe_info->version[0]);

  nodesetprop (curdev, "SIGNATURE", buf);

  /* OEM Strings. */
  nodesetprop (curdev, "OEM_NAME", (char *)vbe_info->oem_name.string);
  if (vbe_info->version[1] >= 3)
  {
    nodesetprop (curdev, "VENDOR", (char *)vbe_info->vendor_name.string);
    nodesetprop (curdev, "PRODUCT", (char *)vbe_info->product_name.string);
    nodesetprop (curdev, "REVISION", (char *)vbe_info->product_revision.string);
  }

  nodesetpropint (curdev, "VIDEOMEMORY", vbe_info->memory_size * 64);
  if (strcasestr (vbe_info->oem_name.string, "intel")
      && strstr (vbe_info->oem_name.string, "810"))
  {
    nodesetprop (curdev, "MEMORY", "4096");
  }

  /* List supported standard modes. */
  mode_list = vbe_info->mode_list.list;
  if (*mode_list != 0xffff)
  {
  for (; *mode_list != 0xffff; mode_list++)
  {
    int i;
    for (i = 0; known_vesa_modes[i].x != 0; i++)
    {
      if (known_vesa_modes[i].number == *mode_list)
      {
        sprintf(buf,"VESA_MODE_%d",i);
        nodesetprop (curdev, buf, (char *)known_vesa_modes[i].text);
      }
    }
  }
  }

  if (vbe_get_edid_supported ())
  {
  edid_info = vbe_get_edid_info ();

  /* Interpret results. */
  if ((edid_info == NULL) || (edid_info->version == 0))
  {
    return devices;
  }

  if ((edid_info->version == 0xff && edid_info->revision == 0xff)
      || (edid_info->version == 0 && edid_info->revision == 0))
  {
    return devices;
  }

  nodesetpropint (curdev, "EDID_VER", edid_info->version);
  nodesetpropint (curdev, "EDID_VER", edid_info->revision);

  manufacturer[0] = edid_info->manufacturer_name.char1 + 'A' - 1;
  manufacturer[1] = edid_info->manufacturer_name.char2 + 'A' - 1;
  manufacturer[2] = edid_info->manufacturer_name.char3 + 'A' - 1;
  manufacturer[3] = '\0';

  nodesetprop (curdev, "MANUFACTURER", manufacturer);
  nodesetpropint (curdev, "EDID_ID", edid_info->product_code);
  sprintf (buf,"%s%04x", manufacturer, edid_info->product_code);
  nodesetprop (curdev, "EISA_ID", buf);

  if (edid_info->serial_number != 0xffffffff)
  {
    if (strcmp (manufacturer, "MAG") == 0)
    {
      edid_info->serial_number -= 0x7000000;
    }
    if (strcmp (manufacturer, "OQI") == 0)
    {
      edid_info->serial_number -= 456150000;
    }
    if (strcmp (manufacturer, "VSC") == 0)
    {
      edid_info->serial_number -= 640000000;
    }
  }
  sprintf (buf,"%08x", edid_info->serial_number);
  nodesetprop (curdev, "EDID_SERIAL_NUMBER", buf);
  nodesetpropint (curdev, "EDID_WEEK", edid_info->week);
  nodesetpropint (curdev, "EDID_YEAR", edid_info->year + 1990);

  nodesetprop (curdev, "SEPARATE_SYNC", edid_info->video_input_definition.separate_sync ? "YES" : "NO");
  nodesetprop (curdev, "COMPOSITE_SYNC", edid_info->video_input_definition.composite_sync ? "YES" : "NO");
  nodesetprop (curdev, "SYNC_ON_GREEN", edid_info->video_input_definition.sync_on_green ? "YES" : "NO");
  nodesetprop (curdev, "SIGNAL", edid_info->video_input_definition.digital ? "digital" : "analog");

  nodesetpropint (curdev, "EDID_SIZE_HORIZ", edid_info->max_size_horizontal);
  nodesetpropint (curdev, "EDID_SIZE_VERT", edid_info->max_size_vertical);

  sprintf (buf,"%f", edid_info->gamma / 100.0 + 1);
  nodesetprop (curdev, "GAMMA", buf);

  nodesetprop (curdev, "RGB", edid_info->feature_support.rgb ? "YES" : "NO");
  nodesetprop (curdev, "ACTIVE_OFF", edid_info->feature_support.active_off ? "NO" : "YES");
  nodesetprop (curdev, "SUSPEND", edid_info->feature_support.suspend ? "NO" : "YES");
  nodesetprop (curdev, "STANDBY", edid_info->feature_support.standby ? "NO" : "YES");

  if (edid_info->established_timings.timing_720x400_70)
    nodesetprop (curdev,"ESTABLISHED_TIMING_1_RES","720x400@70");
  if (edid_info->established_timings.timing_720x400_88)
    nodesetprop (curdev,"ESTABLISHED_TIMING_2_RES","720x400@88");
  if (edid_info->established_timings.timing_640x480_60)
    nodesetprop (curdev,"ESTABLISHED_TIMING_3_RES","640x480@60");
  if (edid_info->established_timings.timing_640x480_67)
    nodesetprop (curdev,"ESTABLISHED_TIMING_4_RES","640x480@67");
  if (edid_info->established_timings.timing_640x480_72)
    nodesetprop (curdev,"ESTABLISHED_TIMING_5_RES","640x480@72");
  if (edid_info->established_timings.timing_640x480_75)
    nodesetprop (curdev,"ESTABLISHED_TIMING_6_RES","640x480@75");
  if (edid_info->established_timings.timing_800x600_56)
    nodesetprop (curdev,"ESTABLISHED_TIMING_7_RES","800x600@56");
  if (edid_info->established_timings.timing_800x600_60)
    nodesetprop (curdev,"ESTABLISHED_TIMING_8_RES","800x600@60");
  if (edid_info->established_timings.timing_800x600_72)
    nodesetprop (curdev,"ESTABLISHED_TIMING_9_RES","800x600@72");
  if (edid_info->established_timings.timing_800x600_75)
    nodesetprop (curdev,"ESTABLISHED_TIMING_10_RES","800x600@75");
  if (edid_info->established_timings.timing_832x624_75)
    nodesetprop (curdev,"ESTABLISHED_TIMING_11_RES","832x624@75");
  if (edid_info->established_timings.timing_1024x768_87i)
    nodesetprop (curdev,"ESTABLISHED_TIMING_12_RES","1024x768@87i");
  if (edid_info->established_timings.timing_1024x768_60)
    nodesetprop (curdev,"ESTABLISHED_TIMING_13_RES","1024x768@60");
  if (edid_info->established_timings.timing_1024x768_70)
    nodesetprop (curdev,"ESTABLISHED_TIMING_14_RES","1024x768@70");
  if (edid_info->established_timings.timing_1024x768_75)
    nodesetprop (curdev,"ESTABLISHED_TIMING_15_RES","1024x768@75");
  if (edid_info->established_timings.timing_1280x1024_75)
    nodesetprop (curdev,"ESTABLISHED_TIMING_16_RES","1280x1024@75");

  /* Standard timings. */
  for (i = 0; i < 8; i++)
  {
    double aspect = 1;
    unsigned int x, y;
    unsigned char xres, vfreq;
    struct vbe_edid_monitor_descriptor *monitor = NULL;
    struct vbe_edid_detailed_timing *timing = NULL;

    xres = edid_info->standard_timing[i].xresolution;
    vfreq = edid_info->standard_timing[i].vfreq;
    if ((xres != vfreq) ||
	((xres != 0) && (xres != 1)) || ((vfreq != 0) && (vfreq != 1)))
    {
      switch (edid_info->standard_timing[i].aspect)
      {
      case 0:
	aspect = 1;
	break;			/*undefined */
      case 1:
	aspect = 0.750;
	break;
      case 2:
	aspect = 0.800;
	break;
      case 3:
	aspect = 0.625;
	break;
      }
      x = (xres + 31) * 8;
      y = x * aspect;

      sprintf (buf,"TIMING_%d_FREQ",i);
      nodesetpropint (curdev, buf, (vfreq & 0x3f) + 60);

      sprintf (buf,"TIMING_%d_RES_HORIZ",i);
      nodesetpropint (curdev, buf, x);

      sprintf (buf,"TIMING_%d_RES_VERT",i);
      nodesetpropint (curdev, buf, y);
    }

  /* Detailed timing information. */
    timing = &edid_info->monitor_details.detailed_timing[i];
    monitor = &edid_info->monitor_details.monitor_descriptor[i];

    if ((monitor->zero_flag_1 != 0) || (monitor->zero_flag_2 != 0))
    {
      sprintf (buf,"TIMING_%d_PIXEL_CLOCK",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_PIXEL_CLOCK (*timing));

      sprintf (buf,"TIMING_%d_HORIZ_ACTIVE_TIME",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_HORIZONTAL_ACTIVE (*timing));

      sprintf (buf,"TIMING_%d_HORIZ_BLANK_TIME",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_HORIZONTAL_BLANKING (*timing));

      sprintf (buf,"TIMING_%d_VERT_ACTIVE_TIME",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_VERTICAL_ACTIVE (*timing));

      sprintf (buf,"TIMING_%d_VERT_BLANK_TIME",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_VERTICAL_BLANKING (*timing));

      sprintf (buf,"TIMING_%d_HORIZ_SYNC_OFFSET",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_HSYNC_OFFSET (*timing));

      sprintf (buf,"TIMING_%d_HORIZ_SYNC_PULSE",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_HSYNC_PULSE_WIDTH (*timing));

      sprintf (buf,"TIMING_%d_VERT_SYNC_OFFSET",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_VSYNC_OFFSET (*timing));

      sprintf (buf,"TIMING_%d_VERT_SYNC_PULSE",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_VSYNC_PULSE_WIDTH (*timing));

      sprintf (buf,"TIMING_%d_HORIZ_SIZE",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_HIMAGE_SIZE (*timing));

      sprintf (buf,"TIMING_%d_VERT_SIZE",i);
      nodesetpropint (curdev, buf, VBE_EDID_DETAILED_TIMING_VIMAGE_SIZE (*timing));
    }
    else if (monitor->type == vbe_edid_monitor_descriptor_serial)
    {
      nodesetpropint (curdev, "MONITOR_DETAILS", i);
      nodesetprop (curdev, "MONITOR_SERIAL", snip (monitor->data.string));
    }
    else if (monitor->type == vbe_edid_monitor_descriptor_ascii)
    {
      nodesetpropint (curdev, "MONITOR_DETAILS", i);
      nodesetprop (curdev, "MONITOR_ASCII", snip (monitor->data.string));
    }
    else if (monitor->type == vbe_edid_monitor_descriptor_name)
    {
      nodesetpropint (curdev, "MONITOR_DETAILS", i);
      nodesetprop (curdev, "MONITOR_NAME", snip (monitor->data.string));
    }
    else if (monitor->type == vbe_edid_monitor_descriptor_range)
    {
      nodesetpropint (curdev, "MONITOR_DETAILS", i);
      nodesetpropint (curdev, "MONITOR_HORIZ_MIN", monitor->data.range_data.horizontal_min);
      nodesetpropint (curdev, "MONITOR_HORIZ_MAX", monitor->data.range_data.horizontal_max);
      nodesetpropint (curdev, "MONITOR_VERT_MIN", monitor->data.range_data.vertical_min);
      nodesetpropint (curdev, "MONITOR_VERT_MAX", monitor->data.range_data.vertical_max);
    }
  }
  }
  return devices;
  }
  }

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
