#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/fd.h>
#include <fcntl.h>
#include <ctype.h>

#include "md5.h"
#include "list.h"
#include "module.h"

static char *
getFloppyDesc (char *name)
{
  int size;
  char *type;
  char desc[64];

  size = atoi (name + 1);

  if (isupper (name[0]))
    type = "3.5\"";
  else
    type = "5.25\"";
  if (size > 1000)
    snprintf (desc, 63, "%s %d.%dMB floppy drive", type, size / 1000,
	      (size % 1000) / 10);
  else
    snprintf (desc, 64, "%s %dKB floppy drive", type, size);
  return strdup (desc);
}

struct node *
probe (struct node *devices, int params)
{
  int fd, x, rc;
  char path[32], name[32],buf[0x100];
  struct floppy_drive_struct ds;
  struct node *newdev;
  MD5_CTX context;
  unsigned char digest[16];

  devices = nodecondremove (devices, "BUS", "FLOPPY", 1);

  for (x = 0; x <= 3; x++)
  {
    snprintf (path, 31, "/dev/fd%d", x);
    fd = open (path, O_RDONLY | O_NONBLOCK);
    if (fd < 0)
      goto cont;
    ioctl (fd, FDRESET, NULL);
    rc = ioctl (fd, FDGETDRVTYP, name);
    if (rc || !name || !strcmp (name, "(null)"))
      goto cont;
    rc = ioctl (fd, FDPOLLDRVSTAT, &ds);
    if (rc)
      goto cont;

    newdev = nodealloc ();
    nodesetprop (newdev, "BUS", "FLOPPY");
    sprintf (buf, "floppy.%d", x);
    nodesetprop (newdev, "CONN", buf);
    nodesetprop (newdev, "CLASS", "FLOPPY");
    nodesetprop (newdev, "DEVICE", path);
    nodesetprop (newdev, "DESC", getFloppyDesc (name));
    nodesetprop (newdev, "DRIVER", "unknown");

    if (ds.track < 0)
      nodesetprop (newdev, "DETACHED", "1");

    MD5Init (&context);
    MD5UpdateS (&context, nodegetprop (newdev, "DESC"));
    MD5Final (digest, &context);

    sprintf (buf,
	     "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
	     digest[0], digest[1], digest[2], digest[3], digest[4],
	     digest[5], digest[6], digest[7], digest[8], digest[9],
	     digest[10], digest[11], digest[12], digest[13],
	     digest[14], digest[15]);

    nodesetprop (newdev, "MD5_ID", buf);
    devices = nodeadd (devices, newdev);

  cont:
    close (fd);
  }

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
