#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "md5.h"
#include "list.h"
#include "module.h"

struct node *
probe (struct node *devices, int params)
{
  DIR *dir;
  struct dirent *ent;
  char path[80];
  char driver[80];
  char media[80];
  char readbuf[256], buf2[0x100];
  char *buf, *ptr;
  int fd, i;
  struct node *newdev;
  MD5_CTX context;
  unsigned char digest[16];

  devices = nodecondremove (devices, "BUS", "IDE", 1);


  if ((dir = opendir ("/proc/ide")))
  {
    while ((ent = readdir (dir)))
    {
      /* if this device is run through ide-scsi, skip it */
      sprintf (path, "/proc/ide/%s/driver", ent->d_name);
      if ((fd = open (path, O_RDONLY)) >= 0)
      {
	i = read (fd, driver, 50);
	close (fd);

	driver[i - 1] = '\0';	/* chop off trailing \n */
      }
      else
      {
	driver[0] = '\0';
      }

      sprintf (path, "/proc/ide/%s/media", ent->d_name);
      if (strncmp (driver, "ide-scsi ", 9)
	  && (fd = open (path, O_RDONLY)) >= 0)
      {
	i = read (fd, media, 50);
	close (fd);
	media[i - 1] = '\0';	/* chop off trailing \n */

	newdev = nodealloc ();
	nodesetprop (newdev, "BUS", "IDE");

	if (!strcmp (media, "cdrom"))
	  nodesetprop (newdev, "CLASS", "CDROM");
	else if (!strcmp (media, "disk"))
	  nodesetprop (newdev, "CLASS", "HD");
	else if (!strcmp (media, "tape"))
	  nodesetprop (newdev, "CLASS", "TAPE");
	else if (!strcmp (media, "floppy"))
	  nodesetprop (newdev, "CLASS", "FLOPPY");
	else
	  nodesetprop (newdev, "CLASS", "OTHER");

	sprintf (buf2, "ide.%s", ent->d_name);
	nodesetprop (newdev, "CONN", buf2);

	sprintf (buf2, "/dev/%s", ent->d_name);
	nodesetprop (newdev, "DEVICE", buf2);
	nodesetprop (newdev, "DRIVER", "ignore");

	sprintf (path, "/proc/ide/%s/model", ent->d_name);
	if ((fd = open (path, O_RDONLY)) >= 0)
	{
	  i = read (fd, path, 50);
	  close (fd);
	  path[i - 1] = '\0';	/* chop off trailing \n */
	  nodesetprop (newdev, "DESC", path);
	}
	else
	{
	  nodesetprop (newdev, "DESC", "Generic IDE device");
	}
	sprintf (path, "/proc/ide/%s/geometry", ent->d_name);
	if ((fd = open (path, O_RDONLY)) >= 0)
	{

	  i = read (fd, readbuf, 255);
	  close (fd);
	  buf = alloca (255);
	  strncpy (buf, readbuf, 254);
	  ptr = buf;
	  while (*buf)
	  {
	    while (*buf && (*buf != '\n'))
	      buf++;
	    if (*buf == '\n')
	    {
	      *buf = '\0';
	      buf++;
	    }
	    if (!strncmp (ptr, "physical", 8))
	      nodesetprop (newdev, "PHYSICAL", ptr + 13);
	    if (!strncmp (ptr, "logical", 7))
	      nodesetprop (newdev, "LOGICAL", ptr + 13);
	    ptr = buf;
	  }
	}

	MD5Init (&context);
	MD5UpdateS (&context, nodegetprop (newdev, "DESC"));
	MD5UpdateS (&context, nodegetprop (newdev, "PHYSICAL"));
	MD5UpdateS (&context, nodegetprop (newdev, "LOGICAL"));
	MD5Final (digest, &context);

	sprintf (buf2,
		 "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
		 digest[0], digest[1], digest[2], digest[3], digest[4],
		 digest[5], digest[6], digest[7], digest[8], digest[9],
		 digest[10], digest[11], digest[12], digest[13], digest[14],
		 digest[15]);
	nodesetprop (newdev, "MD5_ID", buf2);

	devices = nodeadd (devices, newdev);
      }
      errno = 0;
    }
    closedir (dir);
  }
  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *
_init_module (void)
{
  struct modulehandler *mh;
  mh = malloc (sizeof (*mh));
  mh->probe = probe;
  return mh;
}
#endif
