#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/serial.h>
#include <termios.h>
#ifdef __sparc__
#include <asm/kbio.h>
#include <asm/openpromio.h>
#include <asm/types.h>
#endif

#include "md5.h"
#include "list.h"
#include "module.h"

struct speeds
{
	speed_t speed;
	unsigned long value;
};

struct speeds speed_map[] =
{
  {B50, 50},
  {B75, 75},
  {B110, 110},
  {B134, 134},
  {B150, 150},
  {B200, 200},
  {B300, 300},
  {B600, 600},
  {B1200, 1200},
  {B1800, 1800},
  {B2400, 2400},
  {B4800, 4800},
  {B9600, 9600},
  {B19200, 19200},
  {B38400, 38400},
  {B19200, 19200},
  {B38400, 38400},
#ifdef B57600
  {B57600, 57600},
#endif
#ifdef B115200
  {B115200, 115200},
#endif
#ifdef B230400
  {B230400, 230400},
#endif
#ifdef B460800
  {B460800, 460800},
#endif
  {0, 0}
};

#ifdef __sparc__
  static struct
  {
    int layout0, layout1, layout2;
    char *symbols;
    int use_iso9995_3;
  } sunkbd_translate[] = {
  { 0, 33, 80, "",      0, },
  { 1, 34, 81, "",      0, },
  { 2, -1, -1, "fr_BE", 0, },
  { 3, -1, -1, "ca",    0, },
  { 4, 36, 83, "dk",    1, },
  { 5, 37, 84, "de",    1, },
  { 6, 38, 85, "it",    1, },
  { 7, 39, 86, "nl",    0, },
  { 8, 40, 87, "no",    1, },
  { 9, 41, 88, "pt",    1, },
  { 10, 42, 89, "es",   1, },
  { 11, 43, 90, "se",   1, },
  { 12, 44, 91, "fr_CH",1, },
  { 13, 45, 92, "de_CH",1, },
  { 14, 46, 93, "gb",   1, },
  { 16, 47, 94, "ko",   0, },
  { 17, 48, 95, "tw",   0, },
  { 32, 49, 96, "jp",   0, },
  { 50, 97, -1, "fr_CA",0, },
  { 51, -1, -1, "hu",   0, },
  { 52, -1, -1, "pl",   0, },
  { 53, -1, -1, "cs",   0, },
  { 54, -1, -1, "ru",   0, },
  { -1, -1, -1, NULL,   0, }};
#endif

struct node *probe(struct node *devices,int params)
{
  char buf[0x1000];
  MD5_CTX context;
  unsigned char digest[16];
  struct node *newdev;
  int fd;
  char twelve = 12;
  int fdstd = 0;
  devices=nodecondremove(devices,"BUS","KEYBOARD",1);

  newdev=nodealloc();
  nodesetprop(newdev,"BUS","KEYBOARD");
  nodesetprop(newdev,"CONN","kb");
  nodesetprop(newdev,"CLASS","KEYBOARD");

#ifdef __sparc__
  for (fd = 0; fd <= 2; fd++)
  {
    sprintf (buf, "/proc/self/fd/%d", fd);
    if (readlink (buf, buf, 4096) == 12 && !strncmp (buf, "/dev/console", 12))
    {
      fdstd = 1;
      break;
    }
  }
  if (!fdstd)
  {
    fd = open("/dev/console", O_RDWR);
    if (fd < 0)
    {
      newdev=nodefree(newdev);
      return devices;
    }
  }
		
  nodesetprop(newdev,"DRIVER","ignore");
  if (ioctl (fd, TIOCLINUX, &twelve) < 0)
  {
    /* Serial console */
    char desc[64];
    struct serial_struct si;
    int line = 0;

    if (ioctl (fd, TIOCGSERIAL, &si) >= 0)
    {
      if (si.line & 1)
        /* ttyb */
        line = 1;
    }
    if (!fdstd) close(fd);
    sprintf (desc, "Serial console tty%c", line + 'a');
    fd = open("/dev/openprom", O_RDONLY);
    if (fd >= 0)
    {
      struct openpromio *op = (struct openpromio *)buf;
      sprintf (op->oprom_array, "tty%c-mode", line + 'a');
      op->oprom_size = 4096-128-4;
      if (ioctl (fd, OPROMGETOPT, op) >= 0 && op->oprom_size > 0 && op->oprom_size < 40)
      {
        strcat (desc, " ");
	op->oprom_array [op->oprom_size] = 0;
	strcat (desc, op->oprom_array);
      }
      close (fd);
    }
    nodesetprop(newdev,"DESC",desc);
    nodesetprop(newdev,"DEVICE","/dev/console");
  } else
  {
  if (!fdstd) close(fd);
  fd=open("/dev/kbd", O_RDWR);
  if (fd < 0)
  {
    /* PS/2 keyboard */
    kbddev->desc=strdup("Generic PS/2 Keyboard");
  } else
  {
    /* Sun keyboard */
    int kbdtype, kbdlayout, i;
    char *desc, *desclayout = NULL;
    nodesetprop(newdev,"DEVICE","/dev/kbd");
    desc = "Sun Type4 ";
    if (ioctl(fd, KIOCTYPE, &kbdtype) >= 0)
    switch (kbdtype)
    {
      case 2: desc = "Sun Type2 "; break;
      case 3: desc = "Sun Type3 "; break;
      case 4: ioctl(fd, KIOCLAYOUT, &kbdlayout);
      if (kbdlayout < 33)
        desc = "Sun Type4 ";
      else switch (kbdlayout)
      {
        case 33: case 47: case 48: case 49:
        case 80: case 94: case 95: case 96:
          desc = "Sun Type5 "; break;
        case 34: case 81:
          desc = "Sun Type5 Unix "; break;
        default:
          desc = "Sun Type5 Euro "; break;
      }
      for (i = 0; sunkbd_translate[i].layout0 != -1; i++)
      {
        if (sunkbd_translate[i].layout0 == kbdlayout ||
          sunkbd_translate[i].layout1 == kbdlayout ||
          sunkbd_translate[i].layout2 == kbdlayout)
          break;
      }
      if (sunkbd_translate[i].layout0 != -1 && sunkbd_translate[i].symbols[0])
        desclayout = sunkbd_translate[i].symbols;
      break;
    }
    if (desclayout)
    {
      sprintf(buf,"%s %s Keyboard",desc,desclayout);
      nodesetprop(newdev,"DESC",buf);
    } else
    {
      sprintf(buf,"%s Keyboard",desc);
      nodesetprop(newdev,"DESC",buf);
    }
  }
  close (fd);
  }
#else
  for (fd = 0; fd <= 2; fd++)
  {
    sprintf (buf, "/proc/self/fd/%d", fd);
    if (readlink (buf, buf, 4096) == 12 && !strncmp (buf, "/dev/console", 12))
    {
      fdstd = 1;
      break;
    }
  }

  if (!fdstd)
  {
    fd = open("/dev/console", O_RDWR);
    if (fd < 0)
    {
      newdev=nodefree(newdev);
      return devices;
    }
  }

  if (ioctl (fd, TIOCLINUX, &twelve) < 0)
  {
    /* Serial console */
    char desc[64];
    struct serial_struct si;
    struct termios tp;
    speed_t tspeed;
    int line = 0, speed = 0, i;

    if (ioctl (fd, TIOCGSERIAL, &si) >= 0)
    {
      line = si.line;
    }
    if (!tcgetattr(fd, &tp))
    {
      tspeed = cfgetospeed(&tp);
      for (i = 0; speed_map[i].value != 0; i++)
        if (speed_map[i].speed == tspeed)
          speed = speed_map[i].value;
    }
    if (!fdstd) close(fd);
      nodesetprop(newdev,"DRIVER","ignore");
    if (speed)
      sprintf (desc, "Serial console (%d baud) ttyS%d", speed, line);
    else
      sprintf (desc, "Serial console ttyS%d", line);
    nodesetprop(newdev,"DESC",desc);
    sprintf (desc, "/dev/ttyS%d", line);
    nodesetprop(newdev,"DEVICE",desc);
  }
#endif

  MD5Init (&context);
  MD5UpdateS(&context,nodegetprop(newdev,"DESC"));
  MD5Final (digest, &context);
  
  sprintf(buf,"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
    digest[0],digest[1],digest[2],digest[3],
    digest[4],digest[5],digest[6],digest[7],
    digest[8],digest[9],digest[10],digest[11],
    digest[12],digest[13],digest[14],digest[15]);
  nodesetprop(newdev,"MD5_ID",buf);
//  nodesetprop(newdev,"MD5_STATE",buf);
  
  devices=nodeadd(devices,newdev);

  return devices;
}

#ifdef HAPS_MODULES
struct modulehandler *_init_module(void)
{
  struct modulehandler *mh;
  mh=malloc(sizeof(*mh));
  mh->probe=probe;
  return mh;
}
#endif
